/* IR Widget: capture a raw IR signal and dump the timing of the non-demodulated signal

http://www.piclist.com/images/boards/irwidget/index.htm
http://www.hifi-remote.com/forums/dload.php?action=file&file_id=2044
http://www.hifi-remote.com/wiki/index.php?title=IR_Scope_and_IR_Widget_User%27s_Guide
http://www.compendiumarcana.com/irwidget/

Arduino pin numbers for the input capture pin (ICP1) and the logic analyzer debugging pin (LA Dbg):
Board name / MCU                           | ICP1 pin | LA Dbg pin
-------------------------------------------|----------|------------------------
Duemilanove/Uno (ATmega328P / ATmega168)   | 8 (PB0)  |  6 (PD6)
Leonardo (ATmega32U4)                      | 4 (PD4)  | 12 (PD6)
Mega (ATmega2560)                          | - (PD4)  |  - (PD6)

see also here:
http://arduino.cc/en/Hacking/PinMapping168 (also for ATmega328P)
http://arduino.cc/en/Hacking/PinMapping32u4
http://arduino.cc/en/Hacking/PinMapping2560
*/


// Copyright (c) 2012 Michael Dreher  <michael(at)5dot1.de>
// this code may be distributed under the terms of the General Public License V2 (GPL V2)
#include <arduino.h>

////////////////////////////////////////////////////////////////////////////////
// User settings: this section must be adapted to your environment 
////////////////////////////////////////////////////////////////////////////////
#define ARDUINO_SERIAL_BAUDRATE 115200 // adapt this to your serial port baud rate

// the amount of RAM consumed by stack and global variables (e.g. serial receive buffer of 64 byte)
const uint16_t projectRamUsage = 380 + 64;
const uint16_t bufSize ((RAMEND - 0x100 - projectRamUsage) / sizeof(uint16_t)); // use as much RAM as possible
const uint8_t RANGE_EXTENSION_BITS = 4; // factor for upper measurement range = 2^(RANGE_EXTENSION_BITS+1)

// used to debug the timing with a logic analyzer or oszilloscope on a port pin
// Arduino pin numbers of PD6: Duemilanove: pin 6, Leonardo: pin 12
// when you don't need debugging, comment out the #define DEBUG_PIN
//#define DEBUG_PIN 6
//#define DEBUG_PORT D

////////////////////////////////////////////////////////////////////////////////
// Adaption to different MCUs and clk values 
////////////////////////////////////////////////////////////////////////////////
#ifdef defined(_AVR_IOM32U4_H_) || defined(_AVR_IOM2560_H_)
#define CAPTURE_PORT D
#define CAPTURE_PIN 4
#else
// the default is the setting for the ATmega328P
#define CAPTURE_PORT B
#define CAPTURE_PIN 0
#endif

// convert number of clocks to number of nanoseconds, try to use integer arithmetic and avoid
// overflow and too much truncation (double arithmetic costs additional 800 byte of code)
uint32_t inline timerValueToNanoSeconds(uint32_t x)
{
#if (F_CPU % 8000000) == 0
  return ((x * 125UL) / (F_CPU/8000000UL));
#else
#if (F_CPU % 1000000) == 0
  return ((x * 1000UL) / (F_CPU/1000000UL));
#else
#if (F_CPU % 115200) == 0 // serial bps rate compatible cpu clocks, e.g. 7372800 or 14745600
// TODO: this has to be tested, especially the accuracy
  return ((((x * 1000UL) / (F_CPU/115200UL)) * 625UL) / 72UL);
#else
// TODO: this has to be tested
  return (((double)x * 1.0E9) / (double)F_CPU); // use double precision floating point arithmetic
#endif
#endif
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Helper macros
////////////////////////////////////////////////////////////////////////////////
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit)) // clear bit 
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))  // set bit
#define REGNAME2(base, portname) base##portname // internally needed by REGNAME
#define REGNAME(base, portname) REGNAME2(base, portname) // build a register name using a port name

// these macros are used to debug the timing with an logic analyzer or oszilloscope on a port pin
inline void DEBUG_PIN_TOGGLE(void) {
#if defined(DEBUG_PIN) && defined(DEBUG_PORT) 
  REGNAME(PIN, DEBUG_PORT) = _BV(DEBUG_PIN);
#endif
}

inline void DEBUG_PIN_CLEAR(void) {
#if defined(DEBUG_PIN) && defined(DEBUG_PORT) 
  cbi(REGNAME(PORT, DEBUG_PORT), DEBUG_PIN);
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Main part: Capture function
////////////////////////////////////////////////////////////////////////////////
uint16_t captureData[bufSize]; // the buffer where the catured data is stored
int16_t captureCount; // number of values stored in captureData

inline uint16_t packTimeVal(uint16_t diffVal, uint16_t ovlCnt)
{
  // overflow part is stored in the lower RANGE_EXTENSION_BITS bits and not in
  // the upper bits because that makes the code smaller here (less shifting)
  return (0x8000 | ((diffVal >> 1) & (0x7fff & ~(_BV(RANGE_EXTENSION_BITS) - 1))) | ovlCnt);
}  
  
inline uint32_t unpackTimeVal(uint32_t val)
{
  if(val & 0x8000)
  {
    val = val & 0x7fff;
    uint32_t valOvl =  (val & (_BV(RANGE_EXTENSION_BITS) - 1)) << 16;
    uint32_t valTim =  (val << 1) & (0x7fff & ~(_BV(RANGE_EXTENSION_BITS) - 1));
    val = valOvl | valTim;
  }
  
  return val;
}
 
// Wait for a signal on pin ICP1 and store the captured time values in the array 'captureData'
void startCapture(void)
{
  unsigned char sreg = SREG;
  #ifdef USBCON
  // disabling IRQs for a long time will disconnect the USB connection of the ATmega32U4, therefore we
  // defer the sbi() instruction until we got the starting edge and only stop the Timer0 in the meanwhile
  uint8_t tccr0b = TCCR0B;
  TCCR0B &= ~(_BV(CS02) | _BV(CS01) | _BV(CS00)); // stop timer0 (disables timer IRQs)
  #else
  cli(); // disable IRQs
  #endif
  
  register uint8_t ices1_val = _BV(ICES1); 
  register uint8_t tccr1b = TCCR1B | ices1_val; // trigger on rising edge
  TCCR1B = tccr1b;
  OCR1A = TCNT1 - 1;
  TIFR1 = _BV(ICF1) | _BV(OCF1A) | _BV(TOV1); // clear all timer flags
  register uint16_t ovlCnt = 0;
  register uint16_t prevVal = 0;
  register uint8_t tifr; // cache the result of reading TIFR1 (masked with ICF1 and OCF1A)
  register uint16_t *pCapDat; // pointer to current item in captureData[]
  DEBUG_PIN_CLEAR();
  DEBUG_PIN_TOGGLE();
  for(pCapDat = captureData; pCapDat <= &captureData[bufSize - 1]; )
  {
    DEBUG_PIN_TOGGLE();
    // wait for edge or overflow (output compare match)
    while(! (tifr = (TIFR1 & (_BV(ICF1) | _BV(OCF1A))))) {
    }
    DEBUG_PIN_TOGGLE();
    uint16_t val = ICR1;
    OCR1A = val; // timeout based on previous trigger time
 
    if(tifr & _BV(OCF1A)) // check for overflow bit
    {
      if(pCapDat != captureData) // ignore overflow at the beginning of the capture
      {
        ovlCnt++;
        if(ovlCnt >= _BV(RANGE_EXTENSION_BITS))
        {
          TIFR1 = _BV(ICF1) | _BV(OCF1A); // clear input capture and output compare flag bit
          break; // maximum value reached, treat this as timeout and abort capture
        }
      }
      TIFR1 = _BV(ICF1) | _BV(OCF1A); // clear input capture and output compare flag bit
      continue;
    }
    #ifdef USBCON
    cli();
    #endif

    tccr1b ^= ices1_val; // toggle the trigger edge
    TCCR1B = tccr1b;
    TIFR1 = _BV(ICF1) | _BV(OCF1A); // clear input capture and output compare flag bit

    uint16_t diffVal = val - prevVal;
    if(ovlCnt || (diffVal & 0x8000))
    {
      diffVal = packTimeVal(diffVal, ovlCnt);
      ovlCnt = 0;
    }
 
    *pCapDat = diffVal;
    pCapDat++;
    prevVal = val;
  }
  DEBUG_PIN_CLEAR();
 
  // the first array entry contains only the starting time and no time
  // difference, therefore ist is no longer needed and will be removed
  captureCount = (pCapDat - captureData) - 1; // correct count
  for(int16_t i = 0; i < captureCount; i++)
  {
    captureData[i] = captureData[i + 1];
  }

  #ifdef USBCON
  TCCR0B = tccr0b; // re-enable Timer0
  #endif
  SREG = sreg; // enable IRQs
}

////////////////////////////////////////////////////////////////////////////////
// Main loop and output functions
////////////////////////////////////////////////////////////////////////////////
// formatter function: compact format appropraiate for directly reading from screen
inline void dumpBufferCompact()
{
    Serial.print(F("capture["));
    Serial.print(captureCount, DEC);
    Serial.print(F(" values]="));
    Serial.println();
    for(uint16_t i = 0; i < captureCount; i++)
    {
      uint32_t val = timerValueToNanoSeconds(unpackTimeVal(captureData[i]));
      Serial.write((i & 0x01) ? '-' : '+');
      Serial.print(val, DEC);
      if(val >= 50000)
        Serial.println();
      else
        Serial.print(' ');
    }
    Serial.println();
    Serial.println();
}

// formatter function: CSV format to use for OpenOffice or Excel charts
inline void dumpBufferCSV()
{
    Serial.println(F("\"Time [ns]\";\"Signal\";\"Duration\""));
    uint32_t absTime = 0;
    for(uint16_t i = 0; i < captureCount; i++)
    {
      uint32_t val = timerValueToNanoSeconds(unpackTimeVal(captureData[i]));
      Serial.print(absTime, DEC);
      Serial.write(';');
      Serial.write((i & 0x01) ? '1' : '0');
      Serial.write(';');
      Serial.write('0'); // dummy duration
      Serial.println();
      
      Serial.print(absTime + 1, DEC); // lying somewhat to avoid identical timestamps
      Serial.write(';');
      Serial.write((i & 0x01) ? '0' : '1');
      Serial.write(';');
      Serial.println(val, DEC);
      absTime = absTime + val;
    }
    Serial.println();
}

void loop()
{
  startCapture();
  if(captureCount > 0)
  {
    Serial.print(F("bufferSize="));
    Serial.println(bufSize, DEC);

    // select the format you want
    dumpBufferCSV();
    dumpBufferCompact();
  }
}

////////////////////////////////////////////////////////////////////////////////
// Initialization
////////////////////////////////////////////////////////////////////////////////
// initialize Timer 1 and IO pins, needs to be called once before calling startCapture()
void setupCapture()
{
  // configure signal capture pin ICP1 as input (port B0 for ATmega328P)
  cbi(REGNAME(DDR, CAPTURE_PORT), CAPTURE_PIN);

#if defined(DEBUG_PIN) && defined(DEBUG_PORT)
  sbi(REGNAME(DDR, DEBUG_PORT), DEBUG_PIN); // configure logic analyzer debug pin as output
#endif

  // init timer 1
#ifdef PRR0 // for ATmega32U4 (Arduino Leonardo)
  cbi(PRR0, PRTIM1); // disable power save mode of timer
#else
  cbi(PRR, PRTIM1); // disable power save mode of timer
#endif
  TCCR1A = 0; // Timer1 mode 0 = normal
  TCCR1B = _BV(ICNC1) | _BV(CS10); // no prescaler for Timer1, enable noise canceler
}

void setup()
{
  Serial.begin(ARDUINO_SERIAL_BAUDRATE);   // connect to the serial port
  setupCapture();
}

